from typing import Callable, List, Optional

import plotly.graph_objects as go
import pandas as pd
import numpy as np

from insider.mixins.price import PriceIndicatorMixin
from insider.mixins.volume import VolumnIndicatorMixin
from insider.mixins.sar import SARIndicatorMixin
from insider.stock import Stock
from insider.utils import set_layout
from insider.constants import MA_N, MD_N, EXPMA_N, RSI_N, MIKE_COLS, CDP_COLS


class StockInsider(Stock, PriceIndicatorMixin, VolumnIndicatorMixin, SARIndicatorMixin):
    """Plot daily trading indicators."""

    def __init__(self, code, ktype="D"):
        """
        code: Full stock code，(e.g. 'sz002156')，股票完整代码
        ktype: Data frequency, valid input is `D`, `W`, or `M`. 股票数据的频率
        """
        super().__init__(code, ktype)

    @staticmethod
    def _plot_line(df: pd.DataFrame, head: int, line_name: str, y: str = "close"):
        if head:
            df = df.tail(head)
        plot_data = go.Scatter(x=df["day"], y=df[y], name=line_name)
        return plot_data

    def _plot_moving_lines(
        self,
        func: Callable,
        verbose_func: Callable,
        name: str,
        y: str = "close",
        head: int = 90,
        ns: Optional[List] = None,
        verbose: bool = False,
    ):

        plot_data = []
        for n in ns:
            df = func(n=n)
            line_name = name + str(n)
            plot_data.append(self._plot_line(df, head, line_name, y=y))

        if verbose:
            df = self._df.copy()
            verbose_data = verbose_func(df, head)
            plot_data.append(verbose_data)

        layout = set_layout()
        fig = go.Figure(data=plot_data, layout=layout)
        if verbose:
            fig.update_layout(xaxis_rangeslider_visible=False)
        fig.update_layout(title_text=f"{name.upper()} Chart ({self.stock_code})")
        fig.show()

    def _plot(self, df, head, title, lines, verbose: bool = False):
        """General plot functions shared across the class."""
        fig = go.Figure(layout=set_layout())

        if isinstance(lines, str):
            lines = [lines]
        elif not isinstance(lines, list):
            raise ValueError("Only string or list is valid type for lines.")

        for n in lines:
            fig.add_trace(self._plot_line(df, head=head, y=n, line_name=n.upper()))

        if verbose:
            fig.add_trace(self._plot_stock_data(self._df, head))

        fig.update_layout(
            title_text=f"{title} Chart ({self.stock_code})",
            xaxis_rangeslider_visible=False,
        )
        fig.show()

    def plot_ma(
        self, head: int = 90, ns: Optional[List[int]] = None, verbose: bool = False
    ):
        """Plot Moving Average Indicator. 绘出MA曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 5, 10, 20, 60-day lines
            选择曲线的种类，e.g. [5, 10], 默认会绘出5, 10, 20, 60日曲线
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = MA_N

        func = self.ma
        verbose_func = self._plot_stock_data
        self._plot_moving_lines(
            func=func,
            verbose_func=verbose_func,
            name="ma",
            head=head,
            ns=ns,
            verbose=verbose,
        )

    def plot_md(
        self, head: int = 90, ns: Optional[List[int]] = None, verbose: bool = False
    ):
        """Plot Moving Deviation Indicator. 绘出MD曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 5, 10, 20-day lines
            选择曲线的种类，e.g. [5, 10], 默认会绘出5, 10, 20日曲线
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = MD_N

        func = self.md
        verbose_func = self._plot_stock_data
        self._plot_moving_lines(
            func=func,
            verbose_func=verbose_func,
            name="md",
            head=head,
            ns=ns,
            verbose=verbose,
        )

    def plot_ema(
        self, head: int = 90, ns: Optional[List[int]] = None, verbose: bool = False
    ):
        """Plot Exponential Moving Average Indicator. 绘出EMA曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 5, 10, 20, 60-day lines
            选择曲线的种类，e.g. [5, 10, 20, 60], 默认会绘出5, 10, 20, 60日曲线
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = EXPMA_N

        func = self.ema
        verbose_func = self._plot_stock_data
        self._plot_moving_lines(
            func=func,
            verbose_func=verbose_func,
            name="ema",
            head=head,
            ns=ns,
            verbose=verbose,
        )

    def plot_macd(self, head: int = 90):
        """Plot MACD (Moving Average Convergence and Divergence) Indicator. 绘出MACD曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。

        A mixed chart will be plotted, including a bar chart to visualize MACD, and line charts
        to visualize DIFF and DEA.
        将会绘出差值柱形图来表示MACD, 以及表示差离值和讯号线的线性图。
        """
        df_macd = self.macd()
        if head:
            df_macd = df_macd.tail(head)

        df_macd.loc[:, "color"] = df_macd["macd"].apply(
            lambda x: "red" if x >= 0 else "green"
        )

        layout = set_layout()
        fig = go.Figure(layout=layout)

        fig.add_trace(
            go.Bar(
                x=df_macd["day"],
                y=df_macd["macd"],
                base=0,
                marker_color=df_macd["color"],
                name="MACD",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=df_macd["day"], y=df_macd["dea"], marker_color="orange", name="DEA"
            )
        )
        fig.add_trace(
            go.Scatter(
                x=df_macd["day"], y=df_macd["diff"], marker_color="black", name="DIFF"
            )
        )
        fig.update_layout(title_text=f"MACD Chart ({self.stock_code})")
        fig.show()

    def plot_kdj(self, head: int = 90, n: int = 9, smooth_type="sma"):
        """Plot KDJ Indicator. 绘出KDJ曲线。

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 9. 平移平均曲线的窗口大小，默认
            是9个交易日。
            smooth_type: The metric to calculate moving average, default is `sma`, the other
            option is `ema`. 选择计算平移平均曲线的方式，默认是SMA, 另一个选择是EMA。
        """
        df_kdj = self.kdj(n=n, smooth_type=smooth_type)
        if head:
            df_kdj = df_kdj.tail(head)

        plot_data = []
        for col in ["K", "D", "J"]:
            plot_data.append(self._plot_line(df_kdj, head, col, y=col))

        layout = set_layout()
        fig = go.Figure(data=plot_data, layout=layout)
        fig.update_layout(title_text=f"KDJ Chart ({self.stock_code})")
        fig.show()

    def plot_rsi(self, head: int = 90, ns: Optional[List] = None):
        """Plot RSI (Relative Strength Index) Indicator. 绘出RSI曲线。

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 6, 12, 24-day lines
            选择曲线的种类，e.g. [6, 12], 默认会绘出6, 12, 24日曲线
        """
        if ns is None:
            ns = RSI_N

        func = self.rsi
        verbose_func = self._plot_stock_data
        self._plot_moving_lines(
            func=func,
            verbose_func=verbose_func,
            y="rsi",
            name="RSI",
            head=head,
            ns=ns,
            verbose=False,
        )

    def plot_vrsi(self, head: int = 90, ns: Optional[List] = None):
        """Plot VRSI (Volumn Relative Strength Index) Indicator. 绘出VRSI曲线。

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 6, 12, 24-day lines
            选择曲线的种类，e.g. [6, 12], 默认会绘出6, 12, 24日曲线
        """
        if ns is None:
            ns = RSI_N

        func = self.vrsi
        verbose_func = self._plot_stock_data
        self._plot_moving_lines(
            func=func,
            verbose_func=verbose_func,
            y="rsi",
            name="VRSI",
            head=head,
            ns=ns,
            verbose=False,
        )

    @staticmethod
    def _plot_volumn_data(df, head):
        df_volumn = df.copy()
        if head:
            df_volumn = df_volumn.tail(head)

        df_volumn = df_volumn.assign(
            color=lambda x: np.where(x["open"] < x["close"], "red", "green")
        )

        data = go.Bar(
            x=df_volumn["day"],
            y=df_volumn["volumn"],
            base=0,
            marker_color=df_volumn["color"],
            name="Volumn",
        )
        return data

    def plot_volumn(self, head: int = 90):
        """Plot Volumn over time. 绘出交易量能柱状图。

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的交易量能柱状图。
        """
        df_volumn = self._df.copy()
        data = self._plot_volumn_data(df_volumn, head)

        layout = set_layout()
        fig = go.Figure(data=[data], layout=layout)
        fig.update_layout(title_text=f"Volumn Chart ({self.stock_code})")
        fig.show()

    def plot_vma(
        self, head: int = 90, ns: Optional[List] = None, verbose: bool = False
    ):
        """Plot VMA over time. 绘出交易能量MA图曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 5, 10, 20-day lines
            选择曲线的种类，e.g. [5, 10], 默认会绘出5, 10, 20日曲线
            verbose: If to include volumn change bar chart or not, default is False.
            选择是否将能量变化柱状图一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = MA_N

        func = self.vma
        verbose_func = self._plot_volumn_data
        self._plot_moving_lines(
            func=func,
            verbose_func=verbose_func,
            y="volumn",
            name="vma",
            head=head,
            ns=ns,
            verbose=verbose,
        )

    def plot_vmacd(self, head: int = 90):
        """Plot VMACD (Volumn Moving Average Convergence and Divergence)
        Indicator. 绘出VMACD曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。

        A mixed chart will be plotted, including a bar chart to visualize VMACD, and line charts
        to visualize DIFF and DEA.
        将会绘出量能差值柱形图来表示VMACD, 以及表示差离值和讯号线的线性图。
        """
        df_vmacd = self.vmacd()
        if head:
            df_vmacd = df_vmacd.tail(head)

        df_vmacd.loc[:, "color"] = df_vmacd["macd"].apply(
            lambda x: "red" if x >= 0 else "green"
        )

        layout = set_layout()
        fig = go.Figure(layout=layout)

        fig.add_trace(
            go.Bar(
                x=df_vmacd["day"],
                y=df_vmacd["macd"],
                base=0,
                marker_color=df_vmacd["color"],
                name="VMACD",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=df_vmacd["day"], y=df_vmacd["dea"], marker_color="orange", name="DEA"
            )
        )
        fig.add_trace(
            go.Scatter(
                x=df_vmacd["day"], y=df_vmacd["diff"], marker_color="black", name="DIFF"
            )
        )
        fig.update_layout(title_text=f"VMACD Chart ({self.stock_code})")
        fig.show()

    def plot_vstd(self, head: int = 90, ns: Optional[List] = None):
        """Plot VSTD chart. 绘出VSTD曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 5, 10, 20-day lines
            选择曲线的种类，e.g. [5, 10], 默认会绘出5, 10, 20日曲线
        """
        if ns is None:
            ns = MD_N

        func = self.vstd
        verbose_func = self._plot_volumn_data
        self._plot_moving_lines(
            func=func,
            verbose_func=verbose_func,
            y="vstd",
            name="vstd",
            head=head,
            ns=ns,
            verbose=False,
        )

    def plot_env(self, head: int = 90, n: int = 14, verbose: bool = False):
        """Plot ENV indicator. 绘出ENV曲线。

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 14. 平移平均曲线的窗口大小，默认
            是14个交易日。
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """

        df_env = self.env(n=n)
        self._plot(
            df=df_env, head=head, title="ENV", lines=["up", "down"], verbose=verbose
        )

    def plot_vosc(self, head: int = 90):
        """Plot Volumn Oscillator Indicator 绘出成交量震荡指标

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
        """
        df_vosc = self.vosc()
        self._plot(df=df_vosc, head=head, title="VOSC", lines=["vosc"])

    def plot_mi(self, head: int = 90, n: int = 12):
        """Plot Momentum Indicator. 绘出动量指标

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 12. 平移平均曲线的窗口大小，默认
            是12个交易日。
        """
        df_mi = self.mi(n=n)
        self._plot(df=df_mi, head=head, title="MI", lines=["mi"])

    def plot_mike(
        self,
        head: int = 90,
        n: int = 12,
        ns: Optional[List] = None,
        verbose: bool = False,
    ):
        """Plot Mike Base Indicator. 绘出Mike指标

        压力线解释：
        初级压力线（WR)
        中级压力线（MR)
        强力压力线（SR）
        初级支撑线（WS）
        中级支撑线（MS）
        强力支撑线（SS）

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 12. 平移平均曲线的窗口大小，默认
            是12个交易日。
            ns: Choose the lines to be plotted, default is None, which will plot all six lines.
            选择压力线来绘出，默认会绘出所有六条压力线。
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = MIKE_COLS
        elif isinstance(ns, str):
            ns = [ns]
        else:
            ns = [n.lower() for n in ns]
            if [n for n in ns if n not in MIKE_COLS]:
                raise ValueError(
                    "Invalid input for ns, valid values are wr, sr, ws, ms and ss."
                )

        df_mike = self.mike(n=n)
        self._plot(df=df_mike, head=head, title="MIKE", lines=ns, verbose=verbose)

    def plot_adtm(self, head: int = 90):
        """Plot ADTM(23,8) indicator. 绘出动态买卖气指标 (ADTM(23, 8))

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
        """
        df_adtm = self.adtm()
        self._plot(df=df_adtm, head=head, title="ADTM", lines=["adtm", "adtmma"])

    def plot_obv(self, head: int = 90):
        """Plot OBV (On Balance Volumn) Indicator。绘出能量指标

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
        """
        df_obv = self.obv()
        self._plot(df=df_obv, head=head, title="OBV", lines=["obv"])

    def plot_rc(self, head: int = 90, n: int = 30):
        """Plot RC (Price rate of Change) Indicator 绘出价格变化率指标

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 30. 平移平均曲线的窗口大小，默认
            是30个交易日。
        """
        df_rc = self.rc(n=n)
        self._plot(df=df_rc, head=head, title="RC", lines=["arc"])

    def plot_boll(self, head: int = 90, n: int = 26, verbose: bool = False):
        """Plot BOLL line indicator 绘出布林线。

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 26. 平移平均曲线的窗口大小，默认
            是26个交易日。
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        df_boll = self.boll(n=n)
        self._plot(
            df=df_boll,
            head=head,
            title="BOLL",
            lines=["up", "middle", "down"],
            verbose=verbose,
        )

    def plot_bbiboll(
        self, head: int = 90, n: int = 11, m: int = 6, verbose: bool = False
    ):
        """Plot BOLL line indicator 绘出布林线。

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 6. 平移平均曲线的窗口大小，默认
            是6个交易日。
            m: The number to decide the width of up/down lines. 上下压力线的带宽倍数。
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        df_boll = self.bbiboll(n=n, m=m)
        self._plot(
            df=df_boll,
            head=head,
            title="BBIBOLL",
            lines=["upr", "bbiboll", "dwn"],
            verbose=verbose,
        )

    def plot_atr(self, head: int = 90, n: int = 14):
        """Plot Average True Ranger indicator. 绘出真实变化率曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 14. 平移平均曲线的窗口大小，默认
            是14个交易日。
        """
        df_atr = self.atr(n=n)
        self._plot(df=df_atr, head=head, title="ATR", lines=["tr", "atr"])

    def plot_cdp(
        self,
        head: int = 90,
        n: int = 1,
        ns: Optional[List] = None,
        verbose: bool = False,
    ):
        """Plot Contrarian Operation Indicator. 绘出逆势操作曲线

        压力线解释：
        CDP: 需求值
        AH: 最高值
        NH：近高值
        AL：最低值
        NL：近低值

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            n: The size of moving average period for K, default is 1. 平移平均曲线的窗口大小，默认
            是1个交易日。
            ns: Choose the lines to be plotted, default is None, which will plot all five lines.
            选择压力线来绘出，默认会绘出所有五条条压力线。
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = CDP_COLS
        elif isinstance(ns, str):
            ns = [ns]
        else:
            ns = [n.lower() for n in ns]
            if [n for n in ns if n not in CDP_COLS]:
                raise ValueError(
                    "Invalid input for ns, valid values are ah, al, cdp, nh, nl."
                )

        df_cdp = self.cdp(n=n)
        self._plot(df=df_cdp, head=head, title="CDP", lines=ns, verbose=verbose)

    def plot_sar(self, head: int = 90, verbose: bool = False):
        """Plot Stop And Reverse (SAR) indicator. 绘出止损反转指标

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        df_sar = self.sar()

        if head:
            df_sar = df_sar.tail(head)

        fig = go.Figure(layout=set_layout())
        fig.add_trace(
            go.Scatter(
                x=df_sar["day"],
                y=df_sar["sar"],
                marker=dict(color=df_sar["color"], size=3),
                mode="markers",
                name="SAR",
            )
        )

        if verbose:
            fig.add_trace(self._plot_stock_data(self._df, head))

        fig.update_layout(
            title_text=f"SAR Chart ({self.stock_code})",
            xaxis_rangeslider_visible=False,
        )
        fig.show()

    def plot_mtm(self, head: int = 90):
        """Plot MTM indicator. 绘出动量指标

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
        """
        df_mtm = self.mtm()
        self._plot(df=df_mtm, head=head, title="MTM", lines=["mtm", "mtmma"])



import numpy as np

from insider.mixins.base import BaseMixin
from insider.constants import MOVING_VOLUMN_COLS, VOLUMN_VOLS


class VolumnIndicatorMixin(BaseMixin):
    """Volumn Indicator Mixin (量能相关指标混合)"""

    def vma(self, n=5):
        """Volumn Moving Average Calculation (量能移动平均值计算)"""
        df_vma = self._df.loc[:, MOVING_VOLUMN_COLS]
        df_vma.loc[:, "volumn"] = self._ma(col="volumn", n=n)
        return df_vma

    def vmacd(self, n=12, m=26, k=9):
        """Volumn Moving Average Convergence Divergence Calculation (量能平滑异同移动平均计算)"""
        df_vmacd = self._df.loc[:, MOVING_VOLUMN_COLS]
        df_vmacd.loc[:, "diff"] = self._ema(col="volumn", n=n) - self._ema(
            col="volumn", n=m
        )
        df_vmacd.loc[:, "dea"] = self._ema(col="diff", n=k, df=df_vmacd)
        df_vmacd.loc[:, "macd"] = 2 * (df_vmacd["diff"] - df_vmacd["dea"])
        return df_vmacd

    def vstd(self, n=5):
        df_vstd = self._df.loc[:, MOVING_VOLUMN_COLS]
        df_vstd.loc[:, "vstd"] = self._md(col="volumn", n=n, df=df_vstd)
        return df_vstd

    def vrsi(self, n: int = 6):
        return self._rsi("volumn", n=n)

    def vosc(self, n=12, m=26):
        """Volume Oscillator Indicator (成交量震荡指标)

        规则
        SHORT = N周期中成交量的总和/N；
        LONG = M周期中成交量的总和/M；
        VOSC =（SHORT－LONG）÷SHORT×100
        """

        df_vosc = self._df.loc[:, MOVING_VOLUMN_COLS]
        df_vosc.loc[:, "vosc"] = (
            (self._ma(col="volumn", n=n) - self._ma(col="volumn", n=m))
            / self._ma(col="volumn", n=n)
            * 100
        )
        return df_vosc

    def obv(self):
        """On Balance Volumn Indicator (平衡能量指标)

        规则：
        若当日收盘价＞上日收盘价，则当日OBV=前一日OBV＋今日成交量
        若当日收盘价＜上日收盘价，则当日OBV=前一日OBV－今日成交量
        若当日收盘价＝上日收盘价，则当日OBV=前一日OBV
        """
        df_obv = self._df.loc[:, VOLUMN_VOLS]

        df_obv.loc[:, "close_diff"] = df_obv["close"] - df_obv["close"].shift()
        df_obv = df_obv.assign(
            v=lambda x: np.select(
                condlist=[
                    x["close_diff"] > 0,
                    x["close_diff"] < 0,
                    x["close_diff"] == 0,
                ],
                choicelist=[x["volumn"], -x["volumn"], 0],
            )
        )
        df_obv.loc[:, "obv"] = df_obv["v"].expanding(1).sum()
        return df_obv



from typing import Optional
import re

import requests
from requests.exceptions import Timeout
import pandas as pd
import plotly.graph_objects as go

from insider.constants import (
    STOCK_URL,
    KTYPE_CONVERSION,
    KTYPES,
    DAY_COL,
    NUMERIC_COLUMNS,
    MA_COLORS,
    MA_COLS,
)
from insider.utils import set_layout


class Stock:
    """
    Stock Class which collects historical stock trading data and plots the basic
    stock price and k-lines
    """

    def __init__(self, code: str, ktype: str = "D"):
        """
        code: Full stock code，(e.g. 'sz002156')，股票完整代码
        ktype: freq, valid values are `D`, `W`, and `M`，股票趋势频率
        """
        self.code = self._check_code(code)
        self.stock_code = re.findall(r"\d+", self.code)[0]
        self.ktype, self.converted_ktype = self._check_ktype(ktype)
        self.url = STOCK_URL.format(ktype=self.converted_ktype, code=self.code)

        self._df = self._get_stock_data()

    def _check_code(self, code: str) -> str:
        if not code.startswith("sz") and not code.startswith("sh"):
            raise ValueError("Stock code needs to be either sz or sh.")
        elif len(code) != 8:
            raise ValueError(f"Invalid code length: requires 8, but get {len(code)}")
        elif not code[2:].isdigit():
            raise ValueError("Code must be all digits after sh or sz.")
        return code

    def _check_ktype(self, ktype: str) -> (str, str):
        upper_ktype = ktype.upper()
        if upper_ktype not in KTYPES:
            raise ValueError(f"Invalid ktype is given, valid inputs are {KTYPES}")
        converted_ktype = KTYPE_CONVERSION[upper_ktype]
        return upper_ktype, converted_ktype

    @property
    def full_data(self):
        df = self._df.copy()
        return df

    def _get_stock_data(self):
        try:
            r = requests.get(self.url, timeout=10)
        except Timeout:
            raise ValueError("The request timed out. Please try again.")
        else:
            data = r.json()["record"]
            if data:
                df = pd.DataFrame(data, columns=DAY_COL + NUMERIC_COLUMNS)
                df[NUMERIC_COLUMNS] = (
                    df[NUMERIC_COLUMNS]
                    .applymap(lambda x: x.replace(",", ""))
                    .astype("float64")
                )
                self._df = df
                return df
            else:
                raise ValueError(
                    "No data about the stock is found. Please check if the stock code is correct."
                )

    @staticmethod
    def _choose_date(df: pd.DataFrame, start_date: str, end_date: str) -> pd.DataFrame:
        if start_date:
            df = df[df["day"] >= start_date]
        if end_date:
            df = df[df["day"] <= end_date]
        return df

    def show_data(
        self, start_date: Optional[str] = None, end_date: Optional[str] = None
    ):
        """Return the data in Pandas DataFrame，以DataFrame的形式显示历史数据。

        Parameters:
            start_date: start date of data to show, e.g. '2019-01-01'，起始时间
            end_date: end date of data to show, e.g. '2020-01-01'，终止时间

        Returns:
            Truncated DataFrame based on the dates, default is to show
            the full data.
            根据定义的起止时间而截取的历史数据，默认将会返回所有下载的数据。
        """
        df = self._df.copy()
        return self._choose_date(df, start_date, end_date)

    @staticmethod
    def _plot_stock_data(df: pd.DataFrame, head: int):
        if head:
            df = df.tail(head)

        stock_data = go.Candlestick(
            x=df["day"],
            open=df["open"],
            high=df["high"],
            low=df["low"],
            close=df["close"],
            increasing_line_color="red",
            decreasing_line_color="green",
            name="stock price",
        )
        return stock_data

    @staticmethod
    def _plot_ma_data(df: pd.DataFrame, head: int):
        if head:
            df = df.tail(head)

        ma_data = []
        for col, color in zip(MA_COLS, MA_COLORS):
            data = go.Scatter(x=df["day"], y=df[col], name=col, marker_color=color)
            ma_data.append(data)

        return ma_data

    def plot(
        self,
        head: int = 90,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        verbose: bool = True,
    ):
        """Plot the stock price over time. 绘出股票走势图。

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            start_date: start date, default is None, 起始时间
            end_date: end date, default is None, 终止时间
            verbose: If to plot K-line or not, default is True, 是否同时绘出k线，默认是会绘出。
        """
        df = self._df.copy()
        df = self._choose_date(df, start_date, end_date)

        stock_data = self._plot_stock_data(df, head)
        data = [stock_data]
        if verbose:
            ma_data = self._plot_ma_data(df, head)
            data.extend(ma_data)

        fig = go.Figure(data=data, layout=set_layout())
        fig.update_layout(
            xaxis_rangeslider_visible=False,
            title_text=f"Stock Price Chart ({self.stock_code})",
        )
        fig.show()



from typing import Union

import pandas as pd

from insider.constants import VOLUMN_VOLS


class BaseMixin:
    """Base Mixins used in different Indicator Mixins"""

    def _ma(self, col, n, df=None):
        if df is None:
            df = self._df
        return df[col].rolling(n).mean()

    def _md(self, col, n, df=None):
        if df is None:
            df = self._df
        return df[col].rolling(n).std(ddof=0)

    def _ema(self, col, n, df=None):
        if df is None:
            df = self._df
        return df[col].ewm(ignore_na=False, span=n, min_periods=0, adjust=False).mean()

    def _sma(self, n, col=None, df=None, use_ser: Union[bool, pd.Series] = False):
        """Calculate SMA indicator

        if use_ser is assigned with a pandas.Series, then will use it to calculate
        SMA and ignore other given data, default is False which will use the default
        dataset.
        """
        assert n != 0, "Cannot set n to 0 for SMA."

        if df is None:
            df = self._df

        if isinstance(use_ser, pd.Series):
            ser = use_ser.fillna(0)
        elif use_ser is False:
            ser = df[col].fillna(0)
        else:
            raise ValueError("Only False or a Series is allowed for user_ser.")
        return ser.ewm(min_periods=0, ignore_na=False, adjust=False, alpha=1 / n).mean()

    def _rsi(self, col: str, n: int = 6):
        df_rsi = self._df.loc[:, VOLUMN_VOLS]
        ser_shift_diff = df_rsi[col] - df_rsi[col].shift(1)

        df_rsi["shift_diff"] = ser_shift_diff.clip(lower=0)
        df_rsi["shift_diff_abs"] = ser_shift_diff.abs()
        df_rsi["rsi"] = (
            self._sma(col="shift_diff", n=n, df=df_rsi)
            / self._sma(col="shift_diff_abs", n=n, df=df_rsi)
            * 100
        )

        return df_rsi



import numpy as np

from insider.mixins.base import BaseMixin
from insider.constants import MOVING_COLS, HIGH_LOW_COLS, ADTM_COLS


class PriceIndicatorMixin(BaseMixin):
    """Moving Indicator Mixin (移动指标混合)"""

    def ma(self, n=5):
        """Moving Average Calculation (移动平均值计算)"""
        df_ma = self._df.loc[:, MOVING_COLS]
        df_ma.loc[:, "close"] = self._ma(col="close", n=n)
        return df_ma

    def md(self, n=5):
        """Moving Deviation Calculation (移动标准差值计算)"""
        df_md = self._df.loc[:, MOVING_COLS]
        df_md.loc[:, "close"] = self._md(col="close", n=n)
        return df_md

    def ema(self, n=5):
        """Exponential Moving Average Calculation (指数移动平均值计算)"""
        df_ema = self._df.loc[:, MOVING_COLS]
        df_ema.loc[:, "close"] = self._ema(col="close", n=n)
        return df_ema

    def macd(self, n=12, m=26, k=9):
        """Moving Average Convergence Divergence Calculation (平滑异同移动平均计算)

        规则
        DIF = EMA12 - EMA26
        DEA = DIF的9日加权移动平均
        MACD = 2 ×（DIF-DEA）
        """
        df_macd = self._df.loc[:, MOVING_COLS]
        df_macd.loc[:, "diff"] = self._ema(col="close", n=n) - self._ema(
            col="close", n=m
        )
        df_macd.loc[:, "dea"] = self._ema(col="diff", n=k, df=df_macd)
        df_macd.loc[:, "macd"] = 2 * (df_macd["diff"] - df_macd["dea"])
        return df_macd

    def kdj(self, n: int = 9, smooth_type: str = "sma"):
        """
        规则
        n日RSV=（Cn－Ln）/（Hn－Ln）×100
        公式中，Cn为第n日收盘价；Ln为n日内的最低价；Hn为n日内的最高价。
        其次，计算K值与D值：
        * K值 = 2/3×前一日K值 + 1/3×当日RSV
        * D值 = 2/3×前一日D值 + 1/3×当日K值，若无前一日K值与D值，则可分别用50来代替
        * J值 = 3*当日K值 - 2*当日D值
        """
        if smooth_type == "sma":
            func = self._sma
        elif smooth_type == "ema":
            func = self._ema
        else:
            raise ValueError(
                "Invalid smooth average method is given, only sma and ema are allowed."
            )

        df_kdj = self._df.loc[:, HIGH_LOW_COLS]
        close_minus_low = df_kdj["close"] - df_kdj["low"].rolling(n).min()
        high_minus_low = (
            df_kdj["high"].rolling(n).max() - df_kdj["low"].rolling(n).min()
        )

        df_kdj.loc[:, "K"] = (close_minus_low / high_minus_low) * 100
        df_kdj.loc[:, "K"] = func(col="K", n=3, df=df_kdj)
        df_kdj.loc[:, "D"] = func(col="K", n=3, df=df_kdj)
        df_kdj.loc[:, "J"] = 3 * df_kdj["K"] - 2 * df_kdj["D"]

        # Cap it between 0 and 100 as shown in THS.
        df_kdj.loc[:, ["K", "D", "J"]] = df_kdj.loc[:, ["K", "D", "J"]].clip(0, 100)
        return df_kdj

    def rsi(self, n: int = 6):
        return self._rsi("close", n=n)

    def env(self, n: int = 14):
        df_env = self._df.loc[:, MOVING_COLS]

        df_env.loc[:, "up"] = self._ma(col="close", n=n) * 1.06
        df_env.loc[:, "down"] = self._ma(col="close", n=n) * 0.94
        return df_env

    def mi(self, n=12):
        """Calculate MI indicator.
        规则
        MI = CLOSE-REF(CLOSE,1)
        """
        df_mi = self._df.loc[:, MOVING_COLS]

        ser = df_mi["close"] - df_mi["close"].shift(n)
        df_mi.loc[:, "mi"] = self._sma(n=n, use_ser=ser)
        return df_mi

    def mike(self, n: int = 12):
        """Calculate MIKE Base indicator"""
        df_mike = self._df.loc[:, HIGH_LOW_COLS]

        # typ price = avg(high + low + close)
        typ = df_mike[["high", "low", "close"]].mean(axis=1)
        # hv price = highest price in a window
        hv = df_mike["high"].rolling(n).max()
        # lv price = lowest price in a window
        lv = df_mike["low"].rolling(n).min()

        df_mike.loc[:, "wr"] = typ * 2 - lv
        df_mike.loc[:, "mr"] = typ + hv - lv
        df_mike.loc[:, "sr"] = 2 * hv - lv
        df_mike.loc[:, "ws"] = typ * 2 - hv
        df_mike.loc[:, "ms"] = typ - hv + lv
        df_mike.loc[:, "ss"] = 2 * lv - hv

        return df_mike

    def adtm(self, n: int = 23, m: int = 8):
        """动态买卖气指数
        (https://bkso.baidu.com/item/%E5%8A%A8%E6%80%81%E4%B9%B0%E5%8D%96%E6%B0%94%E6%8C%87%E6%A0%87)

        规则：
        1.如果开盘价≤昨日开盘价，DTM=0
          如果开盘价>昨日开盘价，DTM=(最高价-开盘价)和(开盘价-昨日开盘价)的较大值
        2.如果开盘价≥昨日开盘价，DBM=0
          如果开盘价<昨日开盘价，DBM=(开盘价-最低价)和(开盘价-昨日开盘价)的较大值
        3.STM=DTM在N日内的和
        4.SBM=DBM在N日内的和
        5.如果STM>SBM,ADTM=(STM-SBM)/STM
          如果STM<SBM,ADTM=(STM-SBM)/SBM
          如果STM=SBM,ADTM=0
        6.ADTMMA=ADTM的M日简单移动平均
        7.参数N设置为23日，参数M设置为8日

        参考值：
        1.ADTM指标在+1到-1之间波动。
        2.低于-0.5时为低风险区,高于+0.5时为高风险区，需注意风险。
        3.ADTM上穿ADTMMA时，买入股票；ADTM跌穿ADTMMA时，卖出股票。
        """
        df_adtm = self._df.loc[:, ADTM_COLS]

        df_adtm.loc[:, "open_diff"] = df_adtm["open"] - df_adtm["open"].shift(1)
        df_adtm.loc[:, "high_open_diff"] = df_adtm["high"] - df_adtm["open"]
        df_adtm.loc[:, "open_low_diff"] = df_adtm["open"] - df_adtm["low"]

        df_adtm = (
            df_adtm.assign(
                dtm=lambda x: np.where(
                    x["open_diff"] > 0,
                    np.where(
                        x["high_open_diff"] >= x["open_low_diff"],
                        x["high_open_diff"],
                        x["open_low_diff"],
                    ),
                    0,
                )
            )
        ).assign(dbm=lambda x: np.where(x["open_diff"] >= 0, 0, x["open_low_diff"]))

        df_adtm.loc[:, "stm"] = df_adtm["dtm"].rolling(n).sum()
        df_adtm.loc[:, "sbm"] = df_adtm["dbm"].rolling(n).sum()

        df_adtm = df_adtm.assign(
            adtm=lambda x: np.select(
                condlist=[
                    x["stm"] > x["sbm"],
                    x["stm"] < x["sbm"],
                    x["stm"] == x["sbm"],
                ],
                choicelist=[
                    (x["stm"] - x["sbm"]) / x["stm"],
                    (x["stm"] - x["sbm"]) / x["sbm"],
                    0,
                ],
            )
        )

        df_adtm.loc[:, "adtmma"] = self._ma(col="adtm", n=m, df=df_adtm)
        return df_adtm

    def rc(self, n: int = 30):
        """Calculate RC (Price rate of Change) indicator。 计算价格变化率"""
        df_rc = self._df.loc[:, MOVING_COLS]
        df_rc.loc[:, "rc"] = df_rc["close"] / df_rc["close"].shift(n)
        df_rc.loc[:, "arc"] = self._sma(use_ser=df_rc["rc"].shift(1), n=n)

        return df_rc

    def boll(self, n: int = 26):
        """Calculate BOLL line indicator. 计算布林线。

        规则
        中轨线 = N日的移动平均线
        上轨线 = 中轨线 + 两倍的标准差
        下轨线 = 中轨线 － 两倍的标准差
        """
        df_boll = self._df.loc[:, MOVING_COLS]
        df_boll.loc[:, "middle"] = self._ma(col="close", n=n)
        df_boll.loc[:, "up"] = df_boll["middle"] + 2 * self._md(col="close", n=n)
        df_boll.loc[:, "down"] = df_boll["middle"] - 2 * self._md(col="close", n=n)

        return df_boll

    def bbiboll(self, n: int = 11, m: int = 6):
        """Calculate BBIBOLL line indicator. 计算多空布林线.

        规则
        BBIBOLL =（3日简单移动平均价+6日简单移动平均价+12日简单移动平均价+24日简单移动平均价)/4
        UPR = BBIBOLL + M * BBIBOLL的N日估算标准差
        DWN = BBIBOLL - M * BBIBOLL的N日估算标准差
        参数N=11，M=6
        """
        df_bbiboll = self._df.loc[:, MOVING_COLS]
        df_bbiboll.loc[:, "bbiboll"] = (
            self._ma(col="close", n=3)
            + self._ma(col="close", n=6)
            + self._ma(col="close", n=12)
            + self._ma(col="close", n=24)
        ) / 4
        df_bbiboll.loc[:, "upr"] = df_bbiboll["bbiboll"] + m * self._md(
            col="bbiboll", df=df_bbiboll, n=n
        )
        df_bbiboll.loc[:, "dwn"] = df_bbiboll["bbiboll"] - m * self._md(
            col="bbiboll", df=df_bbiboll, n=n
        )
        return df_bbiboll

    def atr(self, n: int = 14):
        """Average True Ranger Indicator."""
        df_atr = self._df.loc[:, HIGH_LOW_COLS]
        df_atr.loc[:, "tr"] = np.vstack(
            [
                (df_atr["high"] - df_atr["low"]).abs(),
                (df_atr["close"].shift(1) - df_atr["high"]).abs(),
                (df_atr["close"].shift(1) - df_atr["low"]).abs(),
            ]
        ).max(axis=0)

        df_atr.loc[:, "atr"] = self._ma(col="tr", df=df_atr, n=n)
        return df_atr

    def cdp(self, n: int = 1):
        """CDP (Contrarian Operation Indicator) 逆势操作指标。

        规则
        需求值（CDP）= (前日最高价+前日最低价+前日收盘价)/3
        最高值（AH）= MA（CDP+（前日最高价-前日最低价），N）
        近高值（NH）= MA（CDP*2-前日最低价，N）
        最低值（AL）= MA（CDP-（前日最高价-前日最低价），N）
        近低值（NL）= MA（CDP*2-前日最高价，N）
        """
        df_cdp = self._df.loc[:, HIGH_LOW_COLS]

        df_cdp.loc[:, "cdp"] = (
            df_cdp.loc[:, ["high", "low", "close"]].shift(1).mean(axis=1)
        )
        df_cdp.loc[:, "ah"] = (
            (df_cdp["cdp"] + df_cdp["high"].shift(1) - df_cdp["low"].shift(1))
            .rolling(n)
            .mean()
        )
        df_cdp.loc[:, "nh"] = (
            (df_cdp["cdp"] * 2 - df_cdp["low"].shift(1)).rolling(n).mean()
        )
        df_cdp.loc[:, "al"] = (
            (df_cdp["cdp"] - df_cdp["high"].shift(1) + df_cdp["low"].shift(1))
            .rolling(n)
            .mean()
        )
        df_cdp.loc[:, "nl"] = (
            (df_cdp["cdp"] * 2 - df_cdp["high"].shift(1)).rolling(n).mean()
        )

        return df_cdp

    def mtm(self, n: int = 6, m: int = 5):
        """Momentum Index. （动量指标）

        规则
        MTM（N日）=C－CN
        C = 当日的收盘价
        CN = N日前的收盘价
        """
        df_mtm = self._df.loc[:, MOVING_COLS]

        df_mtm.loc[:, "mtm"] = df_mtm["close"] - df_mtm["close"].shift(n)
        df_mtm.loc[:, "mtmma"] = self._ma(col="mtm", df=df_mtm, n=m)
        return df_mtm



import pandas as pd

from insider.mixins.base import BaseMixin
from insider.constants import HIGH_LOW_COLS, INITIAL_AF, INITIAL_TREND


class SARIndicatorMixin(BaseMixin):
    def sar(self):
        """
        规则

        先选定一段时间判断为上涨或下跌；
        若是看涨，则第一天的SAR值必须是近期内的最低价；若是看跌，则第一天的SAR须是近期的最高价；
        第二天的SAR，则为第一天的最高价（看涨时）或是最低价（看跌时）与第一天的SAR的差距乘上加速因子，再加上第一天的SAR就可求得；

        每日的SAR都可用上述方法类推，归纳公式如下：
        SAR（n）=SAR（n-1）+AF〖EP（n-1）－SAR（n-1）〗
        SAR（n）=第n日的SAR值，SAR（n-1）即第（n-1）日之值；
        AF：加速因子；
        EP：极点价，若是看涨一段期间，则EP为这段期间的最高价，若是看跌一段时间，则EP为这段期间的最低价；
        EP（n-1）：第（n-1）日的极点价。

        加速因子第一次取0．02，假若第一天的最高价比前一天的最高价还高，则加速因子增加0．02，若无新高则加速因子沿用前一天的数值，
        但加速因子最高不能超过0．2。反之，下跌也类推；

        若是看涨期间，计算出某日的SAR比当日或前一日的最低价高，则应以当日或前一日的最低价为某日之SAR；
        若是看跌期间，计算某日之SAR比当日或前一日的最高价低，则应以当日或前一日的最高价为某日的SAR；
        """
        df_sar = self._df.loc[:, HIGH_LOW_COLS]
        trend = INITIAL_TREND
        af = INITIAL_AF

        df_sar.loc[:, "sar"] = df_sar["close"]
        df_sar.loc[:, "trend"] = pd.Series(index=df_sar["close"].index, dtype="bool")

        start_high = df_sar.loc[0, "high"]
        start_low = df_sar.loc[0, "low"]

        for i in range(1, len(df_sar)):

            reverse = False
            if trend:
                df_sar.loc[i, "sar"] = df_sar.loc[i - 1, "sar"] + (
                    af * (start_high - df_sar.loc[i - 1, "sar"])
                )
                if df_sar.loc[i, "low"] < df_sar.loc[i, "sar"]:
                    reverse = True
                    start_low = df_sar.loc[i, "low"]
                    df_sar.loc[i, "sar"] = start_high
                    af = 0.02
                elif df_sar.loc[i, "high"] > start_high:
                    start_high = df_sar.loc[i, "high"]
                    af = min(af + 0.02, 0.2)
            else:
                df_sar.loc[i, "sar"] = df_sar.loc[i - 1, "sar"] + (
                    af * (start_low - df_sar.loc[i - 1, "sar"])
                )

                if df_sar.loc[i, "high"] > df_sar.loc[i, "sar"]:
                    reverse = True
                    start_high = df_sar.loc[i, "high"]
                    df_sar.loc[i, "sar"] = start_low
                    af = 0.02
                elif df_sar.loc[i, "low"] < start_low:
                    start_low = df_sar.loc[i, "low"]
                    af = min(af + 0.02, 0.2)

            trend = trend != reverse
            df_sar.loc[i, "trend"] = trend

        df_sar.loc[:, "color"] = df_sar["trend"].apply(
            lambda x: "red" if x else "green"
        )
        return df_sar
